/*
 * 
 * Author: Roel Verdult
 * Based on code from: James P. Lynch[1]
 * License: no loyalty, no warranty
 *
 * Summary: Example implementation for OLIMEX MOD-N6610 in combination
 *          with the OLIMEX SAM7-Pxxx (P64/P256) development boards
 *
 * [1] Nokia 6100 LCD Display Driver (Revision 1), James P. Lynch, 2009
 * 
 */
 
#include <inttypes.h>
#include <stdio.h>
#include <string.h>
#include <stdarg.h>

#include "lcd_driver.h"
#include "lcd_font.h"

// 12-bit color definitions 
#define WHITE   0xFFF 
#define BLACK   0x000 
#define RED     0xF00 
#define GREEN   0x0F0 
#define BLUE    0x00F 
#define CYAN    0x0FF 
#define MAGENTA 0xF0F 
#define YELLOW  0xFF0 
#define BROWN   0xB22 
#define ORANGE  0xFA0 
#define PINK    0xF6A 

// GE8 (epson) display commands
#define DISON   0xaf
#define DISOFF  0xae
#define DISNOR  0xa6
#define DISINV  0xa7
#define COMSCN  0xbb
#define DISCTL  0xca
#define SLPIN   0x95
#define SLPOUT  0x94
#define PASET   0x75
#define CASET   0x15
#define DATCTL  0xbc
#define RGBSET8 0xce
#define RAMWR   0x5c
#define RAMRD   0x5d
#define PTLIN   0xa8
#define PTLOUT  0xa9
#define RMWIN   0xe0
#define RMWOUT  0xee
#define ASCSET  0xaa
#define SCSTART 0xab
#define OSCON   0xd1
#define OSCOFF  0xd2
#define PWRCTR  0x20
#define VOLCTR  0x81
#define VOLUP   0xd6
#define VOLDOWN 0xd7
#define TMPGRD  0x82
#define EPCTIN  0xcd
#define EPCOUT  0xcc
#define EPMWR   0xfc
#define EPMRD   0xfd
#define EPSRRD1 0x7c
#define EPSRRD2 0x7d
#define NOP     0x25

// Store LCD display size 
#define LCD_PIXELS 25740

// Pointer to the active SPI port
static AT91PS_SPI pSpi = AT91C_BASE_SPI;

// Pin settings for SPI connection
static const Pin pin_lcd_reset      = { UEXT_PIN3_TX, AT91C_BASE_PIOA, AT91C_ID_PIOA, PIO_OUTPUT_1, PIO_DEFAULT };
static const Pin pin_lcd_backlight  = { UEXT_PIN4_RX, AT91C_BASE_PIOA, AT91C_ID_PIOA, PIO_OUTPUT_1, PIO_DEFAULT };
static const Pin pins_lcd_spi[]     = {
  { UEXT_PIN7_MISO,  AT91C_BASE_PIOA, AT91C_ID_PIOA, PIO_PERIPH_A, PIO_PULLUP  },
  { UEXT_PIN8_MOSI,  AT91C_BASE_PIOA, AT91C_ID_PIOA, PIO_PERIPH_A, PIO_DEFAULT },
  { UEXT_PIN9_SPCK,  AT91C_BASE_PIOA, AT91C_ID_PIOA, PIO_PERIPH_A, PIO_DEFAULT },
  { UEXT_PIN10_NPCS, AT91C_BASE_PIOA, AT91C_ID_PIOA, PIO_PERIPH_A, PIO_DEFAULT },
};

// Buffer for LCD string with End Of String (EOS) character
#define LCD_LINES 14
#define LCD_CHARS 22
static char lcd_lines_buf[LCD_LINES][LCD_CHARS+1];
static uint8_t lcd_lines_start;
static uint8_t lcd_lines_count;

// Counter without precision, but useful for ~ delays 
inline static void lcd_delay(volatile size_t n) { while (n--); }

inline static void WriteSpiCommand(const uint8_t cmd)
{
  SPI_Write(pSpi,UEXT_SPI_NPCS,((uint16_t)cmd) & (~0x0100));
}

inline static void WriteSpiData(const uint8_t data)
{
  SPI_Write(pSpi,UEXT_SPI_NPCS,((uint16_t)data) | 0x0100);
}

void lcd_init_peripherals()
{
  // Configure pin for reset
  PIO_Configure(&pin_lcd_reset,1);

  // Configure pin for backlight
  PIO_Configure(&pin_lcd_backlight,1);

  // Configure pins for SPI communication
  PIO_Configure(pins_lcd_spi,PIO_LISTSIZE(pins_lcd_spi));

  // Master mode, fixed select, disable decoder, FDIV=1 (MCK)
  SPI_Configure(pSpi,AT91C_ID_SPI,AT91C_SPI_MSTR           | // Master Mode
                                  AT91C_SPI_PS_FIXED       | // Fixed Peripheral Select
                                  AT91C_SPI_FDIV           | // Clock Selection
                                  SPI_PCS(UEXT_SPI_NPCS));   // Active Peripheral Chip Selection
 
  SPI_ConfigureNPCS(pSpi,UEXT_SPI_NPCS,(AT91C_SPI_DLYBCT & (0x00 << 24))       | // Delay between Consecutive Transfers (0*32 MCK periods) 
                                       (AT91C_SPI_DLYBS  & (0x00 << 16))       | // Delay Before SPCK (0 MCK period) 
                                       (AT91C_SPI_SCBR   & (0x02 <<  8))       | // Serial Clock Baud Rate (baudrate = MCK/2 = 48054841/2 = 24,027,420 baud 
                                       (AT91C_SPI_BITS   & (AT91C_SPI_BITS_9)) | // Bits per Transfer (9 bits) 
                                       (AT91C_SPI_CSAAT  & (0x00 <<  3))       | // Chip Select Active After Transfer (is active after xfer)
                                       (AT91C_SPI_NCPHA  & (0x00 <<  1))       | // Clock Phase (data captured on falling edge) 
                                       (AT91C_SPI_CPOL   & (0x01 <<  0)));       // Clock Polarity (inactive state is logic one)
  
  // Enable the Serial Peripheral Interface
  SPI_Enable(pSpi);
}

void lcd_clear(uint16_t bColor)
{ 
  size_t i;
  uint8_t PixelRow;
  
  // Row address set  (command 0x2B) 
  WriteSpiCommand(PASET); 
  WriteSpiData(0); 
  WriteSpiData(131); 
  
  // Column address set  (command 0x2A) 
  WriteSpiCommand(CASET); 
  WriteSpiData(0); 
  WriteSpiData(131);

  // Set the display memory to BLACK 
  WriteSpiCommand(RAMWR);
  
  // Blindly dump the background color on the screen
  PixelRow = (bColor & 0xFF);
  for(i = 0; i < LCD_PIXELS; i++) { 
    WriteSpiData(PixelRow); 
  }
  
  // terminate the Write Memory command 
  WriteSpiCommand(NOP);
} 

void lcd_init()
{
  lcd_init_peripherals();
  
  // Hardware reset
  PIO_Clear(&pin_lcd_backlight);
  PIO_Clear(&pin_lcd_reset);
  lcd_delay(1000);
  PIO_Set(&pin_lcd_reset);
  PIO_Set(&pin_lcd_backlight);
  lcd_delay(10000);
	
  WriteSpiCommand(DISCTL);	// Display Control
  WriteSpiData(0x00); // P1: 0x00 = 2 divisions, switching period=8 (default) 
  WriteSpiData(0x20); // P2: 0x20 = nlines/4 - 1 = 132/4 - 1 = 32) 
  WriteSpiData(0x00); // P3: 0x00 = no inversely highlighted lines 
  
	WriteSpiCommand(COMSCN);	// Common Scan
	WriteSpiData(0x01);	// COM1-80: 1->80; COM81-160: 160<-81 (otherwise bottom of screen is upside down)
	WriteSpiCommand(OSCON);	// Oscillation On
  WriteSpiCommand(SLPOUT);	// Sleep Out (exit sleep mode)
  
  WriteSpiCommand(PWRCTR); // Power control
  WriteSpiData(0x0f);   // reference voltage regulator on, circuit voltage follower on, BOOST ON 
  
	WriteSpiCommand(DISINV);	// Inverse Display
  
  WriteSpiCommand(DATCTL);	// Data Control
	WriteSpiData(0x01);	// normal orientation; scan across cols, then rows
	WriteSpiData(0x00);	// RGB arrangement (RGB all rows/cols)
  WriteSpiData(0x02);	// 16 bit-color display
  
  // Electronic Volume Control (brightness/contrast) 
  WriteSpiCommand(VOLCTR); 
  //  WriteSpiData(0x32);   // contrast best values for background (white=0x32, black=0x26) 
  WriteSpiData(0x26); 
  WriteSpiData(3);      // P2 = 3   resistance ratio (only value that works) 
  lcd_delay(10000);
  
	// Clear the screen
  lcd_clear(BLACK);
  
  // Reset the lines count
  lcd_lines_start = 0;
  lcd_lines_count = 0;
  
  // Display On
  WriteSpiCommand(DISON);
}

void lcd_putc(char ch, uint8_t x, uint8_t y, uint16_t fColor, uint16_t bColor)
{ 
  int8_t i,j;
  uint8_t PixelRow; 
  uint8_t Mask; 
  uint16_t Word0; 
  uint16_t Word1; 
  uint8_t* pChar;
  
  // Row address set  (command 0x2B) 
  WriteSpiCommand(PASET); 
  WriteSpiData(x); 
  WriteSpiData(x + FONT_ROW_SIZE - 1); 
  
  // Column address set  (command 0x2A) 
  WriteSpiCommand(CASET); 
  WriteSpiData(y); 
  WriteSpiData(y + FONT_COL_SIZE - 1);
  
  // WRITE MEMORY 
  WriteSpiCommand(RAMWR); 
  
  // Ignore the "fancy" font process for just printing only a space
  if (ch == ' ')
  {
    // Get the background color and print all the pixels
    PixelRow = bColor & 0xFF;
    for(i = 0; i < ((FONT_ROW_SIZE*FONT_COL_SIZE)/2); i++) { 
      WriteSpiData(PixelRow);
      WriteSpiData(PixelRow);
      WriteSpiData(PixelRow);
    }
  } else {
    // get pointer to the last byte of the desired character 
    pChar = ((uint8_t*)font6x8) + (FONT_BYTES_PER_CHAR * (ch - 0x1F)) + FONT_BYTES_PER_CHAR - 1; 

    // loop on each row, working backwards from the bottom to the top 
    for (i = FONT_ROW_SIZE - 1; i >= 0; i--) { 
      
      // copy pixel row from font table and then decrement row 
      PixelRow = *pChar--; 
      
      // loop on each pixel in the row (left to right) 
      // Note: we do two pixels each loop 
      Mask = 0x80; 
      for (j = 0; j < FONT_COL_SIZE; j += 2) { 
        
        // if pixel bit set, use foreground color; else use the background color 
        // now get the pixel color for two successive pixels 
        if ((PixelRow & Mask) == 0) 
          Word0 = bColor; 
        else 
          Word0 = fColor; 
        Mask = Mask >> 1; 
        if ((PixelRow & Mask) == 0) 
          Word1 = bColor; 
        else 
          Word1 = fColor; 
        Mask = Mask >> 1; 
        
        // Optimize, avoid shifts when unnessary 
        if (Word0 == Word1) {
          WriteSpiData(Word0 & 0xFF);
          WriteSpiData(Word0 & 0xFF);
          WriteSpiData(Word0 & 0xFF);
        } else {
          // use this information to output three data bytes 
          WriteSpiData((Word0 >> 4) & 0xFF); 
          WriteSpiData(((Word0 & 0xF) << 4) | ((Word1 >> 8) & 0xF)); 
          WriteSpiData(Word1 & 0xFF);
        }
      }  
    }
  }
  // terminate the Write Memory command 
  WriteSpiCommand(NOP);  
} 

void lcd_puts(char* str, uint8_t x, uint8_t y, uint16_t fColor, uint16_t bColor)
{ 
  // loop until null-terminator is seen 
  while (*str != 0x00)
  { 
    // draw the character 
    lcd_putc(*str++,x,y,fColor,bColor);  
    
    // advance the y position 
    y += 6;
    
    // bail out if y exceeds 131 
    if (y > 131) break; 
  } 

  // Fill the rest of the line with spaces
  while (y < 132)
  { 
    // draw the character 
    lcd_putc(' ',x,y,fColor,bColor);  
    
    // advance the y position 
    y += 6;
  } 
} 

extern signed int vsnprintf();
void lcd_printf(const char *pFormat, ...)
{
  // Declare the line index, count and new  
  uint8_t line_i,line_c,line_n;
  va_list ap;
  
  if (lcd_lines_count == LCD_LINES)
  {
    line_n = lcd_lines_start;
    lcd_lines_start++;
    if (lcd_lines_start == LCD_LINES) lcd_lines_start = 0;
  } else {
    line_n = lcd_lines_count;
    lcd_lines_count++;
  }
  
  // Forward call to vsnprintf
  va_start(ap, pFormat);
  vsnprintf(lcd_lines_buf[line_n],LCD_CHARS,pFormat,ap);
  va_end(ap);
  
  line_i = lcd_lines_start;
  for (line_c=0; line_c<lcd_lines_count; line_c++)
  {
    lcd_puts(lcd_lines_buf[line_i], (line_c*9)+4, 2, WHITE, BLACK);
    line_i++;
    if (line_i == LCD_LINES) line_i = 0;
  }
}

